import json
import os
import logging

logger = logging.getLogger()


def read_service_configuration(configuration_file, section=0):
    """

    :param configuration_file:
    :param section:
    :return:
    """
    try:

        if section:
            service_configuration = json.load(open(configuration_file)).get(section)
        else:
            service_configuration = json.load(open(configuration_file))
        return service_configuration
    except Exception as error:
        logger.error("EXCEPTION opening service configuration:", error)
        return 0


def check_configuration_directory():
    """

    :return:
    """
    service_configuration_directory = os.environ.get('service_configuration_directory', 'configuration')
    if os.path.isdir(service_configuration_directory):
        return os.getcwd() + service_configuration_directory
    else:
        return False


def write_data_to_json_file(data_to_persit, filename):
    """
    this method stores data_to_persit to the given filename

    :param data_to_persit:
    :param filename:
    :return:
    """
    try:
        file_to_open = filename
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(file_to_open))
        with open(file_to_open, 'w') as outfile:
            json.dump(data_to_persit, outfile)
            logger.info(
                'Service Configuration stored to %s\n' % filename)
            return 1
    except Exception as error:
        logger.error(
            'Exception %s occurred writing data to a file %s \n' % (error, data_to_persit))
        return 0



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 05/02/2021
import json
import logging

logger = logging.getLogger()
logger.setLevel(logging.INFO)


def decode_if_it_is_a_valid_json(string):
    try:
        return json.loads(string)
    except Exception as not_json:
        return string


def decode_incoming_msg(message):
    """
    if the message is already a JSON no decoding actions are needed.
    Otherwise the message will be decoded
    :param message:
    :return:
    """
    try:

        if type(message) is not dict:
            message = json.loads(message.decode('ascii'))
        for key in message:
            message[key] = decode_if_it_is_a_valid_json(message[key])
        return message
    except Exception as error:
        logger.error('Exception %s occurred decoding the incoming message %s \n' % (error, message))
        return 0



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 28/01/2021
import os
from pathlib import Path


def build_consumer_configuration_from_os_env_vars(configuration_from_env,group_id, offset='latest'):
    """
    This method builds the configuration for initializing the AVRO Confluent Producer.
    It parses each entry in kwargs and add each key/value to the producer_conf dict by replacing any occurrence of "_" with "."

    :param group_id:
    :param topic:
    :return:
    """
    configuration_from_env['bootstrap_servers'] = configuration_from_env.get('brokers_uri')
    consumer_conf = {
        'enable.auto.commit': 'true',
        'group.id': group_id,
        'default.topic.config': {'auto.offset.reset': offset}}

    if configuration_from_env.get('basic_auth_credentials_source'):
        consumer_conf['schema.registry.basic.auth.credentials.source'] = configuration_from_env.get(
            'basic_auth_credentials_source')
    if configuration_from_env.get('basic_auth_user_info'):
        consumer_conf['schema.registry.basic.auth.user.info'] = configuration_from_env.get('basic_auth_user_info')
    # removing what it's not needed as consumers' config
    configuration_from_env.pop('brokers_uri', '')
    configuration_from_env.pop('basic_auth_credentials_source', '')
    configuration_from_env.pop('basic_auth_user_info', '')

    # translating all the keys replacing the _ with . as requested
    for entry in configuration_from_env:
        if configuration_from_env.get(entry):
            consumer_conf[entry.replace('_', '.')] = configuration_from_env.get(entry)

    if 'ssl.ca.location' not in consumer_conf:
        # here we are sure that the default certificate will be used
        cacert_path = Path(__file__).parent / "../std_ssl_cert/cacert.pem"

        consumer_conf['ssl.ca.location'] = cacert_path
    elif 'ssl.ca.location' in consumer_conf:
        consumer_conf['ssl.ca.location'] = os.path.abspath(consumer_conf.get('ssl.ca.location'))
    if 'ssl.certificate.location' in consumer_conf:
        consumer_conf['ssl.certificate.location'] = os.path.abspath(consumer_conf.get('ssl.certificate.location'))
    if 'ssl.key.location' in consumer_conf:
        consumer_conf['ssl.key.location'] = os.path.abspath(consumer_conf.get('ssl.key.location'))

    if 'security.protocol' not in consumer_conf:
        consumer_conf['security.protocol'] = 'plaintext'
    if configuration_from_env.get('debug') and int(configuration_from_env.get('debug')):
        consumer_conf["debug"] = "cgrp"
    if not int(configuration_from_env.get('api_version_request', 1)):
        # this is for brokers <= 0.10
        consumer_conf['api.version.request'] = 'false'
        consumer_conf['broker.version.fallback'] = '0.9.0.1'

    #sys.stderr.write('[Consumer Conf]:  %s\n' % consumer_conf)

    return consumer_conf



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 04/02/2021
from service_configuration_controller.helpers.operation_with_files import write_data_to_json_file
import unittest
import os
class TestWriteData(unittest.TestCase):

    def test_a_create_dir_and_write_data(self):
        filename = '../../configuration/service_configuration.json'
        configuration_file = os.path.abspath(filename)
        ret = write_data_to_json_file(data_to_persit={"data":"ciao"},filename=configuration_file)
        self.assertEqual(ret,1)

    def test_b_create_dir_and_write_data(self):
        filename = '../../configuration/service_configuration.json'
        configuration_file = os.path.abspath(filename)
        ret = write_data_to_json_file(data_to_persit={"data":"ciao2"},filename=configuration_file)
        self.assertEqual(ret,1)



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 12/01/2021
from confluent_kafka import KafkaError
import os
import datetime
from datetime import datetime, timezone
from service_configuration_controller.helpers.decode_messages import decode_incoming_msg
from service_configuration_controller.helpers.confluent_kafka_configuration_builder import \
    build_consumer_configuration_from_os_env_vars
import time
import logging

logger = logging.getLogger()
logger.setLevel(logging.INFO)
logging.basicConfig(
    format='%(asctime)s:%(message)s',
    level=logging.INFO
)
logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.ERROR)


def request_service_configuration(kafka_producer_instance, monitoring_producer_instance,service_name, initial):
    """
    This method produces a message using the given kafka_producer_instance.
    The purpose of the message is to request a configuration file for the service.
    The timer of 2 seconds needs to avoid sending messages faster than how the consumer can receive a response.

    :param kafka_producer_instance:
    :param service_name:
    :param initial:
    :return:
    """
    if not initial:
        time.sleep(3)

        logger.info(
            'Requesting configuration for service name %s' % service_name)

        kafka_producer_instance.produce_message(value={"service_name": service_name, 'application': 'trialcloud',
                                                       'timestamp': datetime.now(timezone.utc).replace(
                                                           microsecond=0).isoformat()})
        monitor_operation(monitoring_producer_instance, service_name)

        return 1
    else:

        # print('no need')
        return 0


def monitor_operation(monitoring_producer_instance, service_name):
    message = {"service_name": service_name, 'application': 'trialcloud',
               'configuration_requested': 1,
               'configuration_retrieved': 0,
               'timestamp': datetime.now(timezone.utc).replace(
                   microsecond=0).isoformat()}
    monitoring_producer_instance.produce_message(value=message)


def set_configuration(service_conf_to_store, abs_path_to_configuration_directory=None):
    """
    This method writes the given dicti service_conf_to_store to the abs_path_to_configuration_directory
    :param service_conf_to_store:
    :return:
    """

    from service_configuration_controller.helpers.operation_with_files import write_data_to_json_file

    if abs_path_to_configuration_directory is None:
        configuration_file = os.path.abspath('configuration') + '/service_configuration.json'
    else:
        configuration_file = abs_path_to_configuration_directory + '/service_configuration.json'
    return write_data_to_json_file(data_to_persit=service_conf_to_store, filename=configuration_file)


def init_consumer_with_schema_registry(configuration_from_env, group_id, offset):
    """
    This method creates an instance of AvroConsumer
    :param configuration_from_env:
    :param group_id:
    :param offset:
    :return:
    """
    from confluent_kafka.avro import AvroConsumer
    consumer_conf = build_consumer_configuration_from_os_env_vars(configuration_from_env=configuration_from_env,
                                                                  group_id=group_id, offset=offset)
    return AvroConsumer(consumer_conf)


def init_consumer(configuration_from_env, group_id, offset):
    """
    This method creates an instance of Consumer
    :param configuration_from_env:
    :param group_id:
    :param offset:
    :return:
    """
    from confluent_kafka import Consumer
    consumer_conf = build_consumer_configuration_from_os_env_vars(configuration_from_env=configuration_from_env,
                                                                  group_id=group_id, offset=offset)
    return Consumer(**consumer_conf)


class ConfigurationConsumer:

    def __init__(self, **kwargs):
        self.service_name = kwargs.get('service_name')
        self.consumer_topic = kwargs.get('topic')
        consumer_configuration = kwargs.get('configuration_from_env')
        if consumer_configuration.get('schema_registry_url'):
            self.consumer = init_consumer_with_schema_registry(
                configuration_from_env=kwargs.get('configuration_from_env'), group_id=kwargs.get('group_id'),
                offset=kwargs.get('offset'))
        else:
            self.consumer = init_consumer(configuration_from_env=kwargs.get('configuration_from_env'),
                                          group_id=kwargs.get('group_id'), offset=kwargs.get('offset'))

        self.service_conf_requester_instance = kwargs.get('service_conf_requester_instance')
        self.monitoring_producer_instance = kwargs.get('monitoring_producer_instance')
        self.initial = kwargs.get('initial', 0)
        self.just_listen = kwargs.get('just_listen', 0)
        self.lets_start = 0

    def run(self):
        if not self.just_listen:
            msg = 'Services Configuration Tool v2.0.0 [Requesting Configuration Mode Enabled ✔] %s \n' % self.consumer_topic
        elif self.just_listen:
            msg = 'Services Configuration Tool v2.0.0 [Listener Mode Enabled ✔] %s \n' % self.consumer_topic
        logger.info(msg)

        try:
            self.consumer.subscribe(self.consumer_topic, on_assign=self.print_assignment)

            while True:

                incoming_message = self.consumer.poll(1)
                if incoming_message is None:
                    if self.lets_start and not self.just_listen and not request_service_configuration(
                            service_name=self.service_name,
                            initial=self.initial,
                            kafka_producer_instance=self.service_conf_requester_instance,monitoring_producer_instance=self.monitoring_producer_instance):
                        break
                    continue
                elif incoming_message.error() is None:
                    message_payload = decode_incoming_msg(message=incoming_message.value())

                    if message_payload.get('service_name') == self.service_name:
                        message = {"service_name": self.service_name, 'application': 'trialcloud',
                                   'configuration_requested': 1,
                                   'configuration_retrieved': 1,
                                   'timestamp': datetime.now(timezone.utc).replace(
                                       microsecond=0).isoformat()}
                        if set_configuration(service_conf_to_store=message_payload):
                            self.initial = 1
                        else:
                            message['configuration_retrieved'] = 0
                            self.initial = 0
                        self.monitoring_producer_instance.produce_message(value=message)


                elif incoming_message.error():
                    if incoming_message.error().code() == KafkaError._PARTITION_EOF:
                        # End of partition event
                        logger.error('%% %s [%d] reached end at offset %d\n' %
                                     (incoming_message.topic(), incoming_message.partition(),
                                      incoming_message.offset()))
                    elif incoming_message.error():
                        logger.error(
                            '[' + datetime.now(timezone.utc).replace(
                                microsecond=0).isoformat() + '] Consumer Error {incoming_message.error()}\n')
                        continue


        except Exception as whatever_it_is:
            logger.error(
                '[' + datetime.now(timezone.utc).replace(
                    microsecond=0).isoformat() + '] EXCEPTION %s on the consumer for topic(s) %s\n' % (
                    whatever_it_is, self.consumer_topic))
            exit(1)

        finally:
            self.consumer.close()


    def print_assignment(self, consumer, partitions):
        logger.info(
            'Consumer %s Topic:%s  Assignment:%s' % (
                consumer,self.consumer_topic, partitions))
        self.lets_start = 1



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 04/02/2021
from service_configuration_controller.helpers.decode_messages import decode_incoming_msg
import unittest


class TestWriteData(unittest.TestCase):

    def test_a_decode_a_json_message(self):
        message = {'service_name': 'my-service-name', 'timestamp': '2021-02-05T15:15:02+00:00',
                   'start_environment': 'production', 'datacentre': '{}',
                   'external_rest_services': '{"repository_url": "http://repository:3000/api/"}',
                   'persistence_conf': '{}',
                   'credentials': '{"authenticated_user_data": {"id": "su-user@xxx.com", "domain": "xxx", "email": "su-user@xxx.com"}}'}
        decoded_version = {'service_name': 'my-service-name', 'timestamp': '2021-02-05T15:15:02+00:00', 'start_environment': 'production', 'datacentre': {}, 'external_rest_services': {'repository_url': 'http://repository:3000/api/'}, 'persistence_conf': {}, 'credentials': {'authenticated_user_data': {'id': 'su-user@xxx.com', 'domain': 'xxx', 'email': 'su-user@xxx.com'}}}
        ret = decode_incoming_msg(message=message)
        self.assertEqual(ret, decoded_version)

    def test_b_return_the_same_json_version(self):
        decoded_version = {'service_name': 'my-service-name', 'timestamp': '2021-02-05T15:15:02+00:00', 'start_environment': 'production', 'datacentre': {}, 'external_rest_services': {'repository_url': 'http://repository:3000/api/'}, 'persistence_conf': {}, 'credentials': {'authenticated_user_data': {'id': 'su-user@xxx.com', 'domain': 'xxx', 'email': 'su-user@xxx.com'}}}
        ret = decode_incoming_msg(message=decoded_version)
        self.assertEqual(ret, decoded_version)



import setuptools
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='service_configuration_layer',
    version='0.0.7',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['avro-python3','requests', 'confluent-kafka','confluent-kafka-producers-wrapper'],
    url='https://github.com/antoniodimariano/services_configuration_tool',
    license='',
    include_package_data=True,
    python_requires='~=3.7',
    author='Antonio Di Mariano',
    author_email='antonio.dimariano@gmail.com',
    description='A tool to request and store services configuration. This tool is part of the Microservices ToolBox',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



# Created by Antonio Di Mariano (antonio.dimariano@gmail.com) at 12/01/2021
import threading
import os
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from confluent_kafka_producers_wrapper.producer import Producer

logger = logging.getLogger()
default_headers = {"Content-Type": "application/json"}
logging.basicConfig(
    format='%(asctime)s.%(msecs)s:%(name)s:%(thread)d:%(levelname)s:%(process)d:%(message)s',
    level=logging.INFO
)
logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)


def wait_to_retrieve_service_configuration(service_name=os.environ.get('service_name')):
    """
    topic_for_setting_conf=tc-set-service-configuration
    topic_for_getting_conf=tc-get-service-configuration
    :param service_name:
    :return:
    """
    if not service_name:
        raise Exception("You must specify a service_name value either as ENV variable or parameter")
    topic_for_getting_conf = os.environ.get('topic_for_getting_conf', 'tc-get-service-configuration')
    topic_for_setting_conf = os.environ.get('topic_for_setting_conf', 'tc-set-service-configuration')
    topic_for_setting_conf_no_avro = os.environ.get('topic_for_setting_conf_no_avro', 0)
    topic_for_getting_conf_no_avro = os.environ.get('topic_for_getting_conf_no_avro', 0)
    topic_for_monitoring_communication = os.environ.get('monitoring_communication_topic',
                                                        'services-configuration-monitoring')
    brokers_configuration = {
        'service_name': service_name,
        'brokers_uri': os.environ.get('brokers'),
        'schema_registry_url': os.environ.get('schema_registry'),
        'security_protocol': os.environ.get('security_protocol'),
        'ssl_ca_location': os.environ.get('ssl_ca_location'),
        'ssl_certificate_location': os.environ.get('ssl_certificate_location'),
        'ssl_key_location': os.environ.get('ssl_key_location'),
        'basic_auth_credentials_source': os.environ.get('basic_auth_credentials_source'),
        'sasl_mechanisms': os.environ.get('sasl_mechanisms'),
        'debug': os.environ.get('debug'),
        'api_version_request': os.environ.get('api_version_request', 1)

    }

    service_name = brokers_configuration.get('service_name')
    brokers_configuration.pop('service_name')
    if topic_for_getting_conf_no_avro:
        brokers_configuration.pop('schema_registry_url', '')
    monitoring_producer_instance = Producer(brokers_configuration=brokers_configuration,
                                            topic=topic_for_monitoring_communication)
    service_conf_requester_instance = Producer(brokers_configuration=brokers_configuration,
                                               topic=topic_for_getting_conf)

    brokers_configuration['group_id'] = os.environ.get('group_id', 'conf_requester_' + uuid.uuid4().hex)

    __start_consumer_and_producer_and_wait_until_service_conf_is_received(brokers_configuration=brokers_configuration,
                                                                          service_conf_requester_instance=service_conf_requester_instance,
                                                                          service_name=service_name,
                                                                          topic_for_setting_conf=topic_for_setting_conf,
                                                                          monitoring_producer_instance=monitoring_producer_instance)

    return listen_service_configuration()


def listen_service_configuration(service_name=os.environ.get('service_name')):
    """
    topic_for_setting_conf=tc-set-service-configuration
    topic_for_getting_conf=tc-get-service-configuration
    :param service_name:
    :return:
    """
    if not service_name:
        raise Exception("You must specify a service_name value either as ENV variable or parameter")
    topic_for_getting_conf = os.environ.get('topic_for_getting_conf', 'tc-get-service-configuration')
    topic_for_setting_conf = os.environ.get('topic_for_setting_conf', 'tc-set-service-configuration')
    topic_for_setting_conf_no_avro = os.environ.get('topic_for_setting_conf_no_avro', 0)
    topic_for_getting_conf_no_avro = os.environ.get('topic_for_getting_conf_no_avro', 0)
    topic_for_monitoring_communication = os.environ.get('monitoring_communication_topic',
                                                        'services-configuration-monitoring')
    brokers_configuration = {
        'service_name': service_name,
        'brokers_uri': os.environ.get('brokers'),
        'schema_registry_url': os.environ.get('schema_registry'),
        'security_protocol': os.environ.get('security_protocol'),
        'ssl_ca_location': os.environ.get('ssl_ca_location'),
        'ssl_certificate_location': os.environ.get('ssl_certificate_location'),
        'ssl_key_location': os.environ.get('ssl_key_location'),
        'basic_auth_credentials_source': os.environ.get('basic_auth_credentials_source'),
        'sasl_mechanisms': os.environ.get('sasl_mechanisms'),
        'debug': os.environ.get('debug'),
        'api_version_request': os.environ.get('api_version_request', 1)

    }

    service_name = brokers_configuration.get('service_name')
    brokers_configuration.pop('service_name')
    if topic_for_getting_conf_no_avro:
        brokers_configuration.pop('schema_registry_url', '')
    monitoring_producer_instance = Producer(brokers_configuration=brokers_configuration,
                                            topic=topic_for_monitoring_communication)
    service_conf_requester_instance = Producer(brokers_configuration=brokers_configuration,
                                               topic=topic_for_getting_conf)

    brokers_configuration['group_id'] = os.environ.get('group_id', 'conf_requester_' + uuid.uuid4().hex)

    __start_consumer_to_listen_to_future_updates_for_service_configuration(brokers_configuration=brokers_configuration,
                                                                           service_conf_requester_instance=service_conf_requester_instance,
                                                                           service_name=service_name,
                                                                           topic_for_setting_conf=topic_for_setting_conf,
                                                                           monitoring_producer_instance=monitoring_producer_instance)


def __init_consumer(brokers_configuration,
                    service_conf_requester_instance,
                    service_name,
                    topic_for_setting_conf, monitoring_producer_instance, just_listen=0):
    from service_configuration_controller.communication.consumers import ConfigurationConsumer

    consumer = ConfigurationConsumer(
        service_conf_requester_instance=service_conf_requester_instance,
        configuration_from_env=brokers_configuration,
        service_name=service_name,
        group_id=brokers_configuration.get('group_id'), offset=brokers_configuration.get('offset', 'latest'),
        topic=[topic_for_setting_conf],
        monitoring_producer_instance=monitoring_producer_instance,
        just_listen=just_listen)

    return consumer


def __start_consumer_to_listen_to_future_updates_for_service_configuration(brokers_configuration,
                                                                           service_conf_requester_instance,
                                                                           service_name,
                                                                           topic_for_setting_conf,
                                                                           monitoring_producer_instance, just_listen=0):
    """

    Here a consumer is started just listen to the given topic_for_setting_conf and update the configuration if new message will arrive.
    There will no outgoing messages to request the service configuration.
    The Thread is not-blocking and the control will be returned to the caller.

    :param brokers_configuration:
    :param service_conf_requester_instance:
    :param service_name:
    :param topic_for_setting_conf:
    :return:
    """

    consumerb = __init_consumer(brokers_configuration,
                                service_conf_requester_instance,
                                service_name,
                                topic_for_setting_conf, monitoring_producer_instance, just_listen=1)
    consumerb_th = threading.Thread(target=consumerb.run)
    consumerb_th.start()


def __start_consumer_and_producer_and_wait_until_service_conf_is_received(brokers_configuration,
                                                                          service_conf_requester_instance, service_name,
                                                                          topic_for_setting_conf,
                                                                          monitoring_producer_instance):
    """
    This method
    :param brokers_configuration:
    :param service_conf_requester_instance:
    :param service_name:
    :param topic_for_setting_conf:
    :return:
    """

    consumer = __init_consumer(brokers_configuration,
                               service_conf_requester_instance,
                               service_name,
                               topic_for_setting_conf, monitoring_producer_instance)
    consumer_th = threading.Thread(target=consumer.run)
    consumer_th.start()
    consumer_th.join()
